/*
 *  The Apache Software License, Version 1.1
 *
 *  Copyright (c) 2000 The Apache Software Foundation.  All rights
 *  reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 *
 *  3. The end-user documentation included with the redistribution,
 *  if any, must include the following acknowledgment:
 *  "This product includes software developed by the
 *  Apache Software Foundation (http://www.apache.org/)."
 *  Alternately, this acknowledgment may appear in the software itself,
 *  if and wherever such third-party acknowledgments normally appear.
 *
 *  4. The names "Apache" and "Apache Software Foundation" must
 *  not be used to endorse or promote products derived from this
 *  software without prior written permission. For written
 *  permission, please contact apache@apache.org.
 *
 *  5. Products derived from this software may not be called "Apache",
 *  nor may "Apache" appear in their name, without prior written
 *  permission of the Apache Software Foundation.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 *  ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package net.cafebabe.sat.xml;

import java.io.File;
import java.io.IOException;
import java.util.Vector;
import net.cafebabe.sat.util.FileTask;
import net.cafebabe.sat.util.IO;
import org.apache.tools.ant.BuildException;

/**
 * This task merge XML files in a single file (see documentation for more
 * details).
 *
 * @author <a href="mailto:casa@sweetohm.net">Michel CASABIANCA</a>
 */
public class MergeTask
    extends FileTask {
    
    /** The destination file */
    private File tofile;
    /** The resulting document encoding */
    private String encoding;
    /** Guessed resulting document encoding (from first file) */
    private String guessedEncoding;
    /** The resulting document doctype */
    private String doctype;
    /** Default generated file */
    private final static String defaultToFile = "index.xml";
    /** Indicates if a nesting element is necessary */
    private boolean nestingElement = true;
    /** Nesting element name */
    private String nestingElementName = "index";
    /** Force flag, merging is performed even there is no newer file */
    private boolean force;
    
    /**
     * Destination file setter.
     *
     * @param tofile  The destination file.
     */
    public void setTofile(String tofile) {
        this.tofile = getProject().resolveFile(tofile);
    }
    
    /**
     * Encoding setter.
     *
     * @param encoding  The encoding of the XML files
     */
    public void setEncoding(String encoding) {
        this.encoding = encoding;
    }
    
    /**
     * Doctype setter.
     *
     * @param doctype  The doctype of the XML file.
     */
    public void setDoctype(String doctype) {
        this.doctype = doctype;
    }
    
    /**
     * Indicate if a nesting element is necessary.
     *
     * @param nestingElement
     */
    public void setNestingElement(boolean nestingElement) {
        this.nestingElement = nestingElement;
    }
    
    /**
     * Indicate the name of the nesting element.
     *
     * @param nestingElementName
     */
    public void setNestingElementName(String nestingElementName) {
        this.nestingElementName = nestingElementName;
    }
    
    /**
     * Set the force attribute (force merging if true).
     *
     * @param force Force flaf as a boolean
     */
    public void setForve(boolean force) {
        this.force=force;
    }
    
    /**
     * The method to launch the program as an Ant task.
     */
    public void execute() {
        if(tofile!=null && !tofile.getParentFile().exists()) {
            throw new BuildException("Directory " + tofile.getParent() +
                                     " not found.");
        }
        if(tofile==null) {
            tofile=getProject().resolveFile(defaultToFile);
        }
        if(!force && !newerFilesThan(tofile)) return;
        Vector files=getFiles(true);
        if(files.size()==0) return;
        log("Merging " + files.size() + " files into " + tofile + ".");
        StringBuffer body = new StringBuffer();
        if (doctype != null) {
            body.append("<!DOCTYPE " + doctype + ">\n\n");
            body.append(
                        "<" + doctype.substring(0, doctype.indexOf(" ")) +
                        ">\n");
        } else {
            if (nestingElement) {
                body.append("<" + nestingElementName + ">\n");
            }
        }
        try {
            for (int i = 0; i < files.size(); i++) {
                File file = (File)files.elementAt(i);
                body.append(processFile(file));
            }
            if(doctype != null) {
                body.append("\n</"+doctype.substring(0,doctype.indexOf(" "))+">");
            } else {
                if(nestingElement) {
                    body.append("\n</"+nestingElementName+">");
                }
            }
            if(encoding==null) {
                encoding=guessedEncoding;
            }
            StringBuffer document=new
                StringBuffer("<?xml version=\"1.0\" encoding=\"").
                append(encoding).
                append("\"?>\n\n").
                append(body);
            IO.writeFile(document.toString(),tofile,encoding);
        } catch (IOException e) {
            throw new BuildException("IO error: "+e.getMessage());
        }
    }
    
    /**
     * Process a single file
     *
     * @param file             The file to process
     * @return Description of the Return Value
     * @exception IOException  Description of the Exception
     */
    private String processFile(File file)
        throws IOException {
        if (encoding==null && guessedEncoding==null) {
            guessedEncoding = IO.getXMLFileEncoding(file);
        }
        String content = IO.loadXMLFile(file);
        return extract(content);
    }
    
    /**
     * Extract the root element from the file content
     *
     * @param content  The content of the file
     * @return Description of the Return Value
     */
    private String extract(String content) {
        int start=0;
        boolean ok=false;
        do {
            start=content.indexOf("<",start);
            if (content.indexOf("<!--",start)==0) {
                start=content.indexOf("-->",start)+3;
            }
            else if(content.charAt(start+1)=='?' ||
                    content.charAt(start+1)=='!') {
                start=content.indexOf(">",start)+1;
            }
            else {
                ok=true;
            }
        } while(!ok);
        return content.substring(start,content.length());
    }
}
